/* -*- Mode: C++; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-    */
/* ex: set filetype=cpp softtabstop=4 shiftwidth=4 tabstop=4 cindent expandtab: */

/*
  $Id: mtsMedtronicStealthlinkExampleComponent.cpp 3836 2012-09-12 18:43:09Z dmirota1 $

  Author(s):  Peter Kazanzides, Anton Deguet
  Created on: 2011-07-14

  (C) Copyright 2011-2012 Johns Hopkins University (JHU), All Rights Reserved.

--- begin cisst license - do not edit ---

This software is provided "as is" under an open source license, with
no warranty.  The complete license can be found in license.txt and
http://www.cisst.org/cisst/license.txt.

--- end cisst license ---
*/

#include <vtkMatrix4x4.h>
#include <vtkTransform.h>
#include <vtkVertexGlyphFilter.h>
#include <vtkPolyDataMapper.h>
#include <vtkActor.h>
#include <vtkProperty.h>
#include <vtkRenderer.h>
#include <vtkRenderWindow.h>
#include <vtkRenderWindowInteractor.h>
#include <math.h>

#include <cisstMultiTask/mtsInterfaceRequired.h>
#include <cisstMultiTask/mtsTransformationTypes.h>
#include <cisstParameterTypes/prmPositionCartesianGet.h>
#include <cisstNumerical/nmrRegistrationRigid.h>

#ifndef SAW_MEDTRONIC_WITHOUT_STEALTHLINK_TYPES
    #ifdef SAW_MEDTRONIC_STEALTHLINK2
        #include <sawMedtronicStealthlink/mtsMedtronicStealthlink2Types.h>
    #else
        #include <sawMedtronicStealthlink/mtsMedtronicStealthlinkTypes.h>
    #endif
#endif

#include <cisstMultiTask/mtsFixedSizeVectorTypes.h>
#include "mtsMedtronicStealthlinkExampleComponent.h"

mtsMedtronicStealthlinkExampleComponent::mtsMedtronicStealthlinkExampleComponent(const std::string & name,
                                                                                 double periodInSeconds):
    mtsTaskPeriodic(name, periodInSeconds),
    BatchReadyEventCounter(0),
    CollectionRunning(false),
    SamplesCollected(0)
{

	AddStealthlinkInterface();
	AddKeyboardInterface();
	ReadFiducials();
	ReadArtery();
	InitVariables();

}

mtsMedtronicStealthlinkExampleComponent::mtsMedtronicStealthlinkExampleComponent(const std::string & name,
                                                                                 double periodInSeconds,
                                                                                 bool enableStateCollectionInterface):
    mtsTaskPeriodic(name, periodInSeconds),
    BatchReadyEventCounter(0),
    CollectionRunning(false),
    SamplesCollected(0)
{

  AddStealthlinkInterface();
	AddKeyboardInterface();
	ReadFiducials();
	ReadArtery();
	InitVariables();
		
  //state collection
  if(enableStateCollectionInterface)
    AddStateCollectionInterface();
}

void mtsMedtronicStealthlinkExampleComponent::InitVariables()
{
  // Initialize variables for the program, which starts in pre-operating mode
	IsOperating = false;
	IsRegistered = false;
	IsCollecting = false;
	IsCorrected = false;
	NoOfRPointsCollected = 0;
	ShotLaser = false;
	NoOfIntersectFailures = 0;
	origError = -1;
	newError = -1;
	noIntersect = 0;

	frameIntersectPoints = vtkSmartPointer<vtkPoints>::New();	
	registerTrue = vtkSmartPointer<vtkLandmarkTransform>::New();
	transformTrue = vtkSmartPointer<vtkTransformPolyDataFilter>::New();
	bspTree = vtkSmartPointer<vtkModifiedBSPTree>::New();
	introError = vtkSmartPointer<vtkTransformPolyDataFilter>::New();
	registerError = vtkSmartPointer<vtkLandmarkTransform>::New();
	transformError = vtkSmartPointer<vtkTransformPolyDataFilter>::New();
	transformTrueFid = vtkSmartPointer<vtkTransformPolyDataFilter>::New();
	transformErrorFid = vtkSmartPointer<vtkTransformPolyDataFilter>::New();
	transformCorrectedFid = vtkSmartPointer<vtkTransformPolyDataFilter>::New();
	icp = vtkSmartPointer<vtkIterativeClosestPointTransform>::New();
	registerCorrected = vtkSmartPointer<vtkLandmarkTransform>::New();
	transformCorrected = vtkSmartPointer<vtkTransformPolyDataFilter>::New();
}

void mtsMedtronicStealthlinkExampleComponent::ReadFiducials()
{
  // Read in the fiducials 
	fidReader = vtkSmartPointer<vtkSimplePointsReader>::New();
	fidReader->SetFileName("points.txt");
	fidReader->Update();
	ctFidPoints = fidReader->GetOutput()->GetPoints();
	frameFidPoints = vtkSmartPointer<vtkPoints>::New();	
		
  // Introduce some error
	translateFrame = vtkSmartPointer<vtkTransform>::New();
	float tXError = 1.5;
	float tYError = 1.4;
	float tZError = 1.5;
	float rXError = 0.5;
	float rYError = 0.25;
	float rZError = 0.65;
	translateFrame->Translate(tXError, tYError, tZError);
	translateFrame->RotateX(rXError);
	translateFrame->RotateY(rYError);
	translateFrame->RotateZ(rZError);
}

void mtsMedtronicStealthlinkExampleComponent::ReadArtery()
{
  // Read artery as polydata
	surfReader = vtkSmartPointer<vtkGenericDataObjectReader>::New();
	surfReader->SetFileName("artery.vtk");
	surfReader->Update();
	ctArtery = surfReader->GetPolyDataOutput();
}

void mtsMedtronicStealthlinkExampleComponent::AddKeyboardInterface() 
{
	mtsInterfaceRequired * required = AddInterfaceRequired("Keyboard");
	if (required) {
		required->AddEventHandlerVoid(&mtsMedtronicStealthlinkExampleComponent::SwitchPreOp, this, "PreOperation");
		required->AddEventHandlerVoid(&mtsMedtronicStealthlinkExampleComponent::SwitchOp, this, "Operation");
		required->AddEventHandlerVoid(&mtsMedtronicStealthlinkExampleComponent::Register, this, "Register");
		required->AddEventHandlerVoid(&mtsMedtronicStealthlinkExampleComponent::Collect, this, "Fiducial");
		required->AddEventHandlerVoid(&mtsMedtronicStealthlinkExampleComponent::ShootLaser, this, "Laser");
		required->AddEventHandlerVoid(&mtsMedtronicStealthlinkExampleComponent::CorrectRegistration, this, "Correct");
		
	}
}

void mtsMedtronicStealthlinkExampleComponent::AddStealthlinkInterface()
{
    mtsInterfaceRequired * required = AddInterfaceRequired("Stealthlink");
    if (required) {
#ifndef SAW_MEDTRONIC_WITHOUT_STEALTHLINK_TYPES
        required->AddFunction("GetTool", Stealthlink.GetTool);
        required->AddFunction("GetFrame", Stealthlink.GetFrame);
#endif
    }

    // the following two tools are using names normally defined in config.xml
    AddToolInterface("Pointer", Pointer);
    AddToolInterface("Frame", Frame);
}

void mtsMedtronicStealthlinkExampleComponent::AddStateCollectionInterface()
{
    mtsInterfaceRequired * required;

    required = AddInterfaceRequired("CollectorState");
    if (required) {
        required->AddFunction("StartCollection", CollectorState.StartCollection);
        required->AddFunction("StopCollection", CollectorState.StopCollection);
    }
}

void mtsMedtronicStealthlinkExampleComponent::AddToolInterface(const std::string & toolName,
                                                               mtsMedtronicStealthlinkExampleComponent::ToolStruct & functionSet)
{    
	mtsInterfaceRequired * required = AddInterfaceRequired(toolName, MTS_OPTIONAL);
    if (required) {
        required->AddFunction("GetPositionCartesian", functionSet.GetPositionCartesian);
        required->AddFunction("GetHindCartesian", functionSet.GetHindCartesian);
		required->AddFunction("GetMarkerCartesian", functionSet.GetMarkerCartesian);
    }
}

void mtsMedtronicStealthlinkExampleComponent::SwitchPreOp(void)
{
  // Switch to pre-op mode
	// Clear all points collected in op or pre-op mode 
	if (IsOperating) {
		frameIntersectPoints->Reset();
		NoOfIntersects = 0;
		NoOfIntersectFailures = 0;
		IsCorrected = false;
	}
	IsOperating = false;
	IsRegistered = false;
	NoOfRPointsCollected = 0;
	frameFidPoints->Reset();
	std::cout << "In Pre-Op Mode" << std::endl;
}

void mtsMedtronicStealthlinkExampleComponent::SwitchOp(void)
{
  // Only allow switch to operation if already registered
	if (IsRegistered) {
		IsOperating = true;
		std::cout << "Switched to operating mode" << std::endl;
		frameIntersectPoints->Reset();
		NoOfIntersects = 0;
		NoOfIntersectFailures = 0;
		IsCorrected = false;
	}
	else {
		std::cout << "No registration. Cannot switch to operating mode." << std::endl;
	}
}

void mtsMedtronicStealthlinkExampleComponent::ShootLaser(void)
{
  // Only shoot laser if in operating mode
	if (IsOperating) {
		ShotLaser = true;	
	}
	else {
		std::cout << "Not operating. Could not shoot laser." << std::endl;
	}
}

void mtsMedtronicStealthlinkExampleComponent::Register(void)
{
	std::cout << "Registering..." << std::endl;
	if (NoOfRPointsCollected == 6) {
		
		// Get the true registration
		IsRegistered = true;

		std::cout << "Doing a true registration!" << std::endl;
		registerTrue->SetSourceLandmarks(ctFidPoints);
		registerTrue->SetTargetLandmarks(frameFidPoints);
		registerTrue->SetModeToRigidBody();
		registerTrue->Update();

		std::cout << "Doing the true transformation!" << std::endl;
		std::cout << "Get a producer port" << std::endl;
		transformTrue->SetInputConnection(ctArtery->GetProducerPort());
		std::cout << "Try to do the registration" << std::endl;
		transformTrue->SetTransform(registerTrue);
		transformTrue->Update();

		bspTree->SetDataSet(transformTrue->GetOutput());
		bspTree->BuildLocator();

		// Introduce some error
		vtkSmartPointer<vtkPolyData> temp = 
			vtkSmartPointer<vtkPolyData>::New();
		temp->SetPoints(frameFidPoints);
		vtkSmartPointer<vtkVertexGlyphFilter> tempVertexFilter =
			vtkSmartPointer<vtkVertexGlyphFilter>::New();
		tempVertexFilter->SetInputConnection(temp->GetProducerPort());
		tempVertexFilter->Update();

		introError->SetInputConnection(tempVertexFilter->GetOutputPort());
		introError->SetTransform(translateFrame);
		introError->Update();

		// Get the error registration
		std::cout << "Doing the error registration!" << std::endl;
		registerError->SetSourceLandmarks(ctFidPoints);
		registerError->SetTargetLandmarks(introError->GetOutput()->GetPoints());
		registerError->SetModeToRigidBody();
		registerError->Update();

		transformError->SetInputConnection(ctArtery->GetProducerPort());
		transformError->SetTransform(registerError);
		transformError->Update();

		// Calculate average error from ct to error frame coordinates
		vtkSmartPointer<vtkPolyData> trueArtery = transformTrue->GetOutput();
		vtkSmartPointer<vtkPolyData> errorArtery = transformError->GetOutput();
		double truePoint[3];
		double errorPoint[3];
		double averageDistance = 0;
		
		std::cout << "Calculating average error" << std::endl;
		for (vtkIdType i = 0; i < transformTrue->GetOutput()->GetNumberOfPoints(); i++) {
			trueArtery->GetPoint(i, truePoint);
			errorArtery->GetPoint(i, errorPoint);
			averageDistance += sqrt ( (truePoint[0] - errorPoint[0]) * (truePoint[0] - errorPoint[0]) + (truePoint[1] - errorPoint[1]) * (truePoint[1] - errorPoint[1]) + (truePoint[2] - errorPoint[2]) * (truePoint[2] - errorPoint[2]));
		}
		std::cout << "Error introduced: " << averageDistance/transformTrue->GetOutput()->GetPoints()->GetNumberOfPoints() << std::endl;
		origError = averageDistance/transformTrue->GetOutput()->GetPoints()->GetNumberOfPoints();

		// Visualize true artery and error artery
		vtkSmartPointer<vtkPolyDataMapper> trueArteryMapper =
			vtkSmartPointer<vtkPolyDataMapper>::New();
		trueArteryMapper->SetInputConnection(trueArtery->GetProducerPort());
		vtkSmartPointer<vtkActor> trueArteryActor =
			vtkSmartPointer<vtkActor>::New();
		trueArteryActor->SetMapper(trueArteryMapper);
		trueArteryActor->GetProperty()->SetColor(0,0,1);

		vtkSmartPointer<vtkPolyDataMapper> errorArteryMapper =
			vtkSmartPointer<vtkPolyDataMapper>::New();
		errorArteryMapper->SetInputConnection(errorArtery->GetProducerPort());
		vtkSmartPointer<vtkActor> errorArteryActor =
			vtkSmartPointer<vtkActor>::New();
		errorArteryActor->SetMapper(errorArteryMapper);
		errorArteryActor->GetProperty()->SetColor(1,0,0);

		// Visualize true artery and error fiducials

		vtkSmartPointer<vtkVertexGlyphFilter> fidGlyphFilter =
			vtkSmartPointer<vtkVertexGlyphFilter>::New();
		fidGlyphFilter->SetInputConnection(fidReader->GetOutput()->GetProducerPort());
		fidGlyphFilter->Update();

		transformTrueFid->SetInputConnection(fidGlyphFilter->GetOutputPort());
		transformTrueFid->SetTransform(registerTrue);
		transformTrueFid->Update();
		vtkSmartPointer<vtkPolyDataMapper> trueFidMapper =
			vtkSmartPointer<vtkPolyDataMapper>::New();
		trueFidMapper->SetInputConnection(transformTrueFid->GetOutputPort());
		vtkSmartPointer<vtkActor> trueFidActor =
			vtkSmartPointer<vtkActor>::New();
		trueFidActor->SetMapper(trueFidMapper);
		trueFidActor->GetProperty()->SetColor(0,0,1);
		trueFidActor->GetProperty()->SetPointSize(4);

		transformErrorFid->SetInputConnection(fidGlyphFilter->GetOutputPort());
		transformErrorFid->SetTransform(registerError);
		transformErrorFid->Update();
		vtkSmartPointer<vtkPolyDataMapper> errorFidMapper =
			vtkSmartPointer<vtkPolyDataMapper>::New();
		errorFidMapper->SetInputConnection(transformErrorFid->GetOutputPort());
		vtkSmartPointer<vtkActor> errorFidActor =
			vtkSmartPointer<vtkActor>::New();
		errorFidActor->SetMapper(errorFidMapper);
		errorFidActor->GetProperty()->SetColor(1,0,0);
		errorFidActor->GetProperty()->SetPointSize(4);

		vtkSmartPointer<vtkRenderer> renderer =
			vtkSmartPointer<vtkRenderer>::New();
		vtkSmartPointer<vtkRenderWindow> renderWindow =
			vtkSmartPointer<vtkRenderWindow>::New();
		renderWindow->AddRenderer(renderer);
		vtkSmartPointer<vtkRenderWindowInteractor> renderWindowInteractor =
			vtkSmartPointer<vtkRenderWindowInteractor>::New();
		renderWindowInteractor->SetRenderWindow(renderWindow);

		renderer->AddActor(trueArteryActor);
		renderer->AddActor(trueFidActor);
		renderer->AddActor(errorArteryActor);
		renderer->AddActor(errorFidActor);
		renderer->SetBackground(.3, .6, .3);

		renderWindow->Render();
		renderWindowInteractor->Start();

	} else {
		std::cout << "Not enough points to register." << std::endl;
	}
}

void mtsMedtronicStealthlinkExampleComponent::Collect(void)
{
	std::cout << "Turned On Collecting..." << std::endl;
	if (! IsOperating)
    IsCollecting = true;
}

void mtsMedtronicStealthlinkExampleComponent::CorrectRegistration(void)
{
	if (! IsOperating)
		return;
	if (! IsRegistered)
		return;
	std::cout << "Trying to correct the registration..." << std::endl;

	vtkSmartPointer<vtkPolyData> temp = 
		vtkSmartPointer<vtkPolyData>::New();
	temp->SetPoints(frameIntersectPoints);

	vtkSmartPointer<vtkVertexGlyphFilter> intersectVertexFilter =
		vtkSmartPointer<vtkVertexGlyphFilter>::New();
	intersectVertexFilter->SetInputConnection(temp->GetProducerPort());
	intersectVertexFilter->Update();

	vtkSmartPointer<vtkPolyData> icpSource =
		vtkSmartPointer<vtkPolyData>::New();
	icpSource->ShallowCopy(intersectVertexFilter->GetOutput());

	icp->SetSource(icpSource);
	icp->SetTarget(transformError->GetOutput());
	icp->GetLandmarkTransform()->SetModeToRigidBody();
	icp->Modified();
	icp->Update();

	// If success
	// Set up transformation
	vtkSmartPointer<vtkPoints> ctNewPoints =
		vtkSmartPointer<vtkPoints>::New();
	ctNewPoints->DeepCopy(ctFidPoints);
	vtkSmartPointer<vtkPoints> frameNewPoints =
		vtkSmartPointer<vtkPoints>::New();
	frameNewPoints->DeepCopy(introError->GetOutput()->GetPoints());

	double newPoint[3];
	for (vtkIdType i = 0; i < NoOfIntersects; i++) {
		icpSource->GetPoint(i, newPoint);
		frameNewPoints->InsertNextPoint(newPoint);
		icp->GetLandmarkTransform()->GetTargetLandmarks()->GetPoint(i, newPoint);
		ctArtery->GetPoint(transformError->GetOutput()->FindPoint(newPoint), newPoint);
		ctNewPoints->InsertNextPoint(newPoint);
	}

	std::cout << "Done setting up transformation..." << std::endl;
	
  // Find the corrected registration
	ctNewPoints->Print(std::cout);
	frameNewPoints->Print(std::cout);
	registerCorrected->SetSourceLandmarks(ctNewPoints);
	registerCorrected->SetTargetLandmarks(frameNewPoints);
	registerCorrected->SetModeToRigidBody();
	registerCorrected->Update();
	registerCorrected->Print(std::cout);
	registerCorrected->GetMatrix()->Print(std::cout);
	
	transformCorrected->SetInputConnection(surfReader->GetOutputPort());
	transformCorrected->SetTransform(registerCorrected);
	transformCorrected->Update();

	vtkSmartPointer<vtkPolyData> trueArtery = transformTrue->GetOutput();
	vtkSmartPointer<vtkPolyData> correctedArtery = transformCorrected->GetOutput();

	double truePoint[3];
	double correctedPoint[3];
	double averageDistance = 0;
	
	std::cout << "Calculating average corrected" << std::endl;
	for (vtkIdType i = 0; i < transformTrue->GetOutput()->GetNumberOfPoints(); i++) {
		trueArtery->GetPoint(i, truePoint);
		correctedArtery->GetPoint(i, correctedPoint);
		averageDistance += sqrt ( (truePoint[0] - correctedPoint[0]) * (truePoint[0] - correctedPoint[0]) + (truePoint[1] - correctedPoint[1]) * (truePoint[1] - correctedPoint[1]) + (truePoint[2] - correctedPoint[2]) * (truePoint[2] - correctedPoint[2]));
	}
	std::cout << "Corrected error: " << averageDistance/transformTrue->GetOutput()->GetPoints()->GetNumberOfPoints() << std::endl;

	newError = averageDistance/transformTrue->GetOutput()->GetPoints()->GetNumberOfPoints();

	// Visualize true artery and error artery
	vtkSmartPointer<vtkPolyDataMapper> trueArteryMapper =
		vtkSmartPointer<vtkPolyDataMapper>::New();
	trueArteryMapper->SetInputConnection(trueArtery->GetProducerPort());
	vtkSmartPointer<vtkActor> trueArteryActor =
		vtkSmartPointer<vtkActor>::New();
	trueArteryActor->SetMapper(trueArteryMapper);
	trueArteryActor->GetProperty()->SetColor(0,0,1);

	vtkSmartPointer<vtkPolyDataMapper> correctedArteryMapper =
		vtkSmartPointer<vtkPolyDataMapper>::New();
	correctedArteryMapper->SetInputConnection(correctedArtery->GetProducerPort());
	vtkSmartPointer<vtkActor> correctedArteryActor =
		vtkSmartPointer<vtkActor>::New();
	correctedArteryActor->SetMapper(correctedArteryMapper);
	correctedArteryActor->GetProperty()->SetColor(1,1,0);
	
	vtkSmartPointer<vtkPolyDataMapper> errorArteryMapper =
		vtkSmartPointer<vtkPolyDataMapper>::New();
	errorArteryMapper->SetInputConnection(transformError->GetOutput()->GetProducerPort());
	vtkSmartPointer<vtkActor> errorArteryActor =
		vtkSmartPointer<vtkActor>::New();
	errorArteryActor->SetMapper(errorArteryMapper);
	errorArteryActor->GetProperty()->SetColor(1,0,0);
	
	vtkSmartPointer<vtkVertexGlyphFilter> fidGlyphFilter =
		vtkSmartPointer<vtkVertexGlyphFilter>::New();
	fidGlyphFilter->SetInputConnection(fidReader->GetOutput()->GetProducerPort());
	fidGlyphFilter->Update();

	transformTrueFid->SetInputConnection(fidGlyphFilter->GetOutputPort());
	transformTrueFid->SetTransform(registerTrue);
	transformTrueFid->Update();
	vtkSmartPointer<vtkPolyDataMapper> trueFidMapper =
		vtkSmartPointer<vtkPolyDataMapper>::New();
	trueFidMapper->SetInputConnection(transformTrueFid->GetOutputPort());
	vtkSmartPointer<vtkActor> trueFidActor =
		vtkSmartPointer<vtkActor>::New();
	trueFidActor->SetMapper(trueFidMapper);
	trueFidActor->GetProperty()->SetColor(0,0,1);
	trueFidActor->GetProperty()->SetPointSize(4);

	transformErrorFid->SetInputConnection(fidGlyphFilter->GetOutputPort());
	transformErrorFid->SetTransform(registerError);
	transformErrorFid->Update();
	vtkSmartPointer<vtkPolyDataMapper> errorFidMapper =
		vtkSmartPointer<vtkPolyDataMapper>::New();
	errorFidMapper->SetInputConnection(transformErrorFid->GetOutputPort());
	vtkSmartPointer<vtkActor> errorFidActor =
		vtkSmartPointer<vtkActor>::New();
	errorFidActor->SetMapper(errorFidMapper);
	errorFidActor->GetProperty()->SetColor(1,0,0);
	errorFidActor->GetProperty()->SetPointSize(4);

	transformCorrectedFid->SetInputConnection(fidGlyphFilter->GetOutputPort());
	transformCorrectedFid->SetTransform(registerCorrected);
	transformCorrectedFid->Update();
	vtkSmartPointer<vtkPolyDataMapper> correctedFidMapper =
		vtkSmartPointer<vtkPolyDataMapper>::New();
	correctedFidMapper->SetInputConnection(transformCorrectedFid->GetOutputPort());
	vtkSmartPointer<vtkActor> correctedFidActor =
		vtkSmartPointer<vtkActor>::New();
	correctedFidActor->SetMapper(correctedFidMapper);
	correctedFidActor->GetProperty()->SetColor(1,1,0);
	correctedFidActor->GetProperty()->SetPointSize(4);

	vtkSmartPointer<vtkRenderer> renderer =
		vtkSmartPointer<vtkRenderer>::New();
	vtkSmartPointer<vtkRenderWindow> renderWindow =
		vtkSmartPointer<vtkRenderWindow>::New();
	renderWindow->AddRenderer(renderer);
	vtkSmartPointer<vtkRenderWindowInteractor> renderWindowInteractor =
		vtkSmartPointer<vtkRenderWindowInteractor>::New();
	renderWindowInteractor->SetRenderWindow(renderWindow);

	renderer->AddActor(trueArteryActor);
	renderer->AddActor(correctedArteryActor);
	renderer->AddActor(errorArteryActor);
	renderer->AddActor(trueFidActor);
	renderer->AddActor(correctedFidActor);
	renderer->AddActor(errorFidActor);
	renderer->SetBackground(.3, .6, .3);

	renderWindow->Render();
	renderWindowInteractor->Start();

	IsCorrected = true;

}

void mtsMedtronicStealthlinkExampleComponent::Run(void)
{
	std::cout << "Operating? " << IsOperating << "\tIsRegistered? " << IsRegistered << "\tIsCorrected? " << IsCorrected << std::endl;

	if (! IsRegistered) {
		std::cout << "Number of points collected: " << NoOfRPointsCollected << std::endl;
	} 
	if (IsOperating) {
		std::cout << "Intersects: " << NoOfIntersects << "\tMisses: " << NoOfIntersectFailures << std::endl;
		std::cout << "Introduced error: " << origError << std::endl;
	}
	if (IsCorrected) {
		std::cout << "Introduced error: " << origError << "\tNew error: " << newError << std::endl;
	}
		
	ProcessQueuedEvents();
	ProcessQueuedCommands();

	
    prmPositionCartesianGet prmPos;
    mtsFrm3 mtsFrm;
    mtsDouble predictedAccuracy;

	prmPositionCartesianGet prmPosPointer, prmPosHind, prmPosFrame;

    bool debug = false;
    bool didOutput = false;
    bool getValid = false;
	mtsExecutionResult result;
	vctFrm3 inverse;
     
    if (Pointer.GetPositionCartesian.IsValid()) {
        result = Pointer.GetPositionCartesian(prmPosPointer);
        if (!result.IsOK()) {
            std::cout << "Pointer.GetPositionCartesian() failed: " << result << std::endl;
        }
        if (prmPosPointer.Valid()) {
            if(debug) {
                std::cout << "Interface Pointer: " << prmPosPointer.Position().Translation() << std::endl;
                std::cout << "Pointer Rotation: " << prmPosPointer.Position().Rotation() << std::endl;
			}
            didOutput = true;		
        }
        result = Pointer.GetHindCartesian(prmPosHind);
        if (!result.IsOK()) {
            std::cout << "Pointer.GetHindCartesian() failed: " << result << std::endl;
        }
        if (prmPosHind.Valid()) {
            if(debug) {
                std::cout << "Interface Pointer: " << prmPosHind.Position().Translation() << std::endl;
                std::cout << "Pointer Rotation: " << prmPosHind.Position().Rotation() << std::endl;
			}
            didOutput = true;
        }
    }

    if (Frame.GetMarkerCartesian.IsValid()) {
        result = Frame.GetMarkerCartesian(prmPosFrame);
        if (!result.IsOK()) {
            std::cout << "Frame.GetMarkerCartesian() failed: " << result << std::endl;
        }
        if (prmPosFrame.Valid()) {
            if(debug) {
                std::cout << "Interface FrameM: " << prmPosFrame.Position().Translation() << std::endl;
                std::cout << "FrameM Rotation: " << prmPosFrame.Position().Rotation() << std::endl;
			}
            didOutput = true;
        } else {
            if(debug)
                std::cerr << "Interface Frame, invalid position" << std::endl;
            didOutput = true;
        }
    }

	if (prmPosPointer.Valid() && prmPosHind.Valid() 
		                      && prmPosFrame.Valid()) {
		inverse = inverse.InverseOf(prmPosFrame.Position());
		
		if (IsCollecting) {
      // Collect a point
			IsCollecting = false;
			vct3 p = inverse * prmPosPointer.Position().Translation();
		    point[0] = p[0];
			point[1] = p[1];
			point[2] = p[2];
			frameFidPoints->InsertNextPoint(point);
			std::cout << "Collected a point!" << std::endl;
			NoOfRPointsCollected = NoOfRPointsCollected + 1;
		}
		if (ShotLaser) {
      // Shoot laser and calculate intersection
			ShotLaser = false;
			vct3 tip = inverse * prmPosPointer.Position().Translation();
			vct3 hind = inverse * prmPosHind.Position().Translation();
			double gradient[3];
			gradient[0] = tip[0] - hind[0];
			gradient[1] = tip[1] - hind[1];
			gradient[2] = tip[2] - hind[2];
			double start[3];
			start[0] = tip[0];
			start[1] = tip[1];
			start[2] = tip[2];
			double end[3];
			end[0] = tip[0] + gradient[0]*10;
			end[1] = tip[1] + gradient[1]*10;
			end[2] = tip[2] + gradient[2]*10;

			double tolerance = .001;
			double t;
			double x[3];
			double pcoords[3];
			int subId;

      // Calculate the intersection
			vtkIdType iD = bspTree->IntersectWithLine(start, end, tolerance, t, x, pcoords, subId);
			std::cout << "Tried an intersect" << std::endl;
			if (iD != noIntersect) {
				NoOfIntersects++;
				frameIntersectPoints->InsertNextPoint(x);
			} else {
				NoOfIntersectFailures++;
			} 
		}
	}
	IsCollecting = false;
	ShotLaser = false;
	
    

}
