#include <vtkVersion.h>
#include <vtkPoints.h>
#include <vtkSmartPointer.h>
#include <vtkLandmarkTransform.h>
#include <vtkMatrix4x4.h>
#include <vtkPolyDataMapper.h>
#include <vtkActor.h>
#include <vtkRenderWindow.h>
#include <vtkRenderer.h>
#include <vtkRenderWindowInteractor.h>
#include <vtkProperty.h>
#include <vtkTransformPolyDataFilter.h>
#include <vtkVertexGlyphFilter.h>
#include <vtkSimplePointsReader.h>
#include <vtkGenericDataObjectReader.h>
#include <vtkTransform.h>
#include <vtkDistancePolyDataFilter.h>
#include <vtkVectorNorm.h>
#include <vtkPointData.h>
#include <vtkCleanPolyData.h>
#include <vtkDataArray.h>
#include <vtkIterativeClosestPointTransform.h>
#include <math.h>
#include <stdlib.h>
#include <time.h>

int run(float reportError[])
{
 
  // Read in fiducials in ct coordinates
  vtkSmartPointer<vtkSimplePointsReader> ctReader =
    vtkSmartPointer<vtkSimplePointsReader>::New();
  ctReader->SetFileName("ct.txt");
  ctReader->Update();

  // Read in fiducials in frame coordinates
  vtkSmartPointer<vtkSimplePointsReader> frameReader =
    vtkSmartPointer<vtkSimplePointsReader>::New();
  frameReader->SetFileName("frame.txt");
  frameReader->Update();

  vtkSmartPointer<vtkPoints> ctPoints = ctReader->GetOutput()->GetPoints();
  vtkSmartPointer<vtkPoints> framePoints = frameReader->GetOutput()->GetPoints();

  // Read in artery in ct coordinates
  vtkSmartPointer<vtkGenericDataObjectReader> arteryReader =
    vtkSmartPointer<vtkGenericDataObjectReader>::New();
  arteryReader->SetFileName("artery.vtk");
  arteryReader->Update();
  vtkPolyData* artery = arteryReader->GetPolyDataOutput();

  // Get the registration from ct to frame coordinates
  vtkSmartPointer<vtkLandmarkTransform> registerTrue = 
    vtkSmartPointer<vtkLandmarkTransform>::New();
  registerTrue->SetSourceLandmarks(ctPoints);
  registerTrue->SetTargetLandmarks(framePoints);
  registerTrue->SetModeToRigidBody();
  registerTrue->Update();
  // registerTrue->GetMatrix()->Print(std::cout);

  // Display the transformation matrix that was computed
  vtkMatrix4x4* mat = registerTrue->GetMatrix();
  // std::cout << "Matrix: " << *mat << std::endl;

  // Transform artery surface from ct to frame coordinates 
  vtkSmartPointer<vtkTransformPolyDataFilter> transformTrue =
    vtkSmartPointer<vtkTransformPolyDataFilter>::New();
  transformTrue->SetInputConnection(arteryReader->GetOutputPort());
  transformTrue->SetTransform(registerTrue);
  transformTrue->Update();

  // Introduce some error to points in frame.txt
  vtkSmartPointer<vtkTransform> translateFrame =
    vtkSmartPointer<vtkTransform>::New();
  // ########### INITIAL TRANSLATIONAL ERROR OF 5
  float introError = 2.0;
  float XError = 2.0;
  float YError = 0.0;
  float ZError = 0.0;
  translateFrame->Translate(0,introError,0);
  translateFrame->RotateX(XError);
  translateFrame->RotateY(YError);
  translateFrame->RotateZ(ZError);
  vtkSmartPointer<vtkTransformPolyDataFilter> someError =
    vtkSmartPointer<vtkTransformPolyDataFilter>::New();
  someError->SetInputConnection(frameReader->GetOutputPort());
  someError->SetTransform(translateFrame);
  someError->Update();

  // Get the error registration from ct to error frame coordinates
  vtkSmartPointer<vtkLandmarkTransform> registerError =
    vtkSmartPointer<vtkLandmarkTransform>::New();
  registerError->SetSourceLandmarks(ctPoints);
  registerError->SetTargetLandmarks(someError->GetOutput()->GetPoints());
  registerError->SetModeToRigidBody();
  registerError->Update();
  // registerError->GetMatrix()->Print(std::cout);
  
  // Transform artery surface from ct to error frame coordinates 
  vtkSmartPointer<vtkTransformPolyDataFilter> transformError =
    vtkSmartPointer<vtkTransformPolyDataFilter>::New();
  transformError->SetInputConnection(arteryReader->GetOutputPort());
  transformError->SetTransform(registerError);
  transformError->Update();

  // Calculate average error from ct to error frame coordinates
  vtkSmartPointer<vtkPolyData> trueArtery = transformTrue->GetOutput();
  vtkSmartPointer<vtkPolyData> errorArtery = transformError->GetOutput();
  double truePoint[3];
  double errorPoint[3];
  double averageDistance = 0;
  for (vtkIdType i = 0; i < transformTrue->GetOutput()->GetPoints()->GetNumberOfPoints(); 
       i++) {
    trueArtery->GetPoint(i, truePoint);
    errorArtery->GetPoint(i, errorPoint);
    averageDistance += sqrt( (truePoint[0]-errorPoint[0])*(truePoint[0]-errorPoint[0]) +
                             (truePoint[1]-errorPoint[1])*(truePoint[1]-errorPoint[1]) +
                             (truePoint[2]-errorPoint[2])*(truePoint[2]-errorPoint[2]) );
    
  }
  float errorError = averageDistance/transformTrue->GetOutput()->GetPoints()->GetNumberOfPoints();
  std::cout << errorError << std::endl;

  // Randomly sample points from ground truth artery surface 

  // ################# CHANGE THIS
  vtkIdType noOfPoints = 35;

  vtkSmartPointer<vtkPoints> samplePoints =
    vtkSmartPointer<vtkPoints>::New();
  vtkPoints * sourcePoints = trueArtery->GetPoints();
  
  double firstPoint[3];
  vtkIdType pid = rand() % trueArtery->GetPoints()->GetNumberOfPoints();
  trueArtery->GetPoint(pid, firstPoint);
  samplePoints->InsertNextPoint(firstPoint);
  int pointsCollected = 1;

  // #################### CHANGE THIS
  int minDistance = 0;
  int maxDistance = 90;
  double randomPoint[3];
  float distance;
  while (pointsCollected < noOfPoints) {
    vtkIdType pid = rand() % trueArtery->GetPoints()->GetNumberOfPoints();
    trueArtery->GetPoint(pid, randomPoint);
    // calculate distance 
    distance = sqrt( (randomPoint[0]-firstPoint[0])*(randomPoint[0]-firstPoint[0]) +
                     (randomPoint[1]-firstPoint[1])*(randomPoint[1]-firstPoint[1]) +
                     (randomPoint[2]-firstPoint[2])*(randomPoint[2]-firstPoint[2]) );
    if (distance < maxDistance && distance >= minDistance) {
      samplePoints->InsertNextPoint(randomPoint);
      pointsCollected++;
    }

  }
  // std::cout << "Number of sample points is " << samplePoints->GetNumberOfPoints() << std::endl;
  
  
  // Calculate the icp 

  vtkSmartPointer<vtkPolyData> temp =
    vtkSmartPointer<vtkPolyData>::New();
  temp->SetPoints(samplePoints);

  vtkSmartPointer<vtkVertexGlyphFilter> sampleVertexFilter =
    vtkSmartPointer<vtkVertexGlyphFilter>::New();
  sampleVertexFilter->SetInputConnection(temp->GetProducerPort());
  sampleVertexFilter->Update();

  vtkSmartPointer<vtkPolyData> icpSource =
    vtkSmartPointer<vtkPolyData>::New();
  icpSource->ShallowCopy(sampleVertexFilter->GetOutput());

  
  vtkSmartPointer<vtkIterativeClosestPointTransform> icp =
    vtkSmartPointer<vtkIterativeClosestPointTransform>::New();
  
  icp->CheckMeanDistanceOn();
  
  // icp from correct surface to ct
  /*
  // START ICP TO CT
  icp->SetSource(icpSource);
  icp->SetMaximumNumberOfIterations(50);
  icp->SetTarget(arteryReader->GetPolyDataOutput()); 
  icp->GetLandmarkTransform()->SetModeToRigidBody();
  icp->Modified();
  icp->Update();
  std::cout << "The icp matrix is: " << *icp->GetMatrix() << std::endl; 
 
  vtkSmartPointer<vtkPoints> ctNewPoints =
    vtkSmartPointer<vtkPoints>::New();
  ctNewPoints->DeepCopy(ctReader->GetOutput()->GetPoints());
  vtkSmartPointer<vtkPoints> frameNewPoints =
    vtkSmartPointer<vtkPoints>::New();
  frameNewPoints->DeepCopy(someError->GetOutput()->GetPoints());
  double newPoint[3];

  // if success, SourceLandmarks == TargetLandmarks
  for (vtkIdType i = 0; i < noOfPoints; i++) {
    icpSource->GetPoint(i, newPoint);
    frameNewPoints->InsertNextPoint(newPoint);
    icp->GetLandmarkTransform()->GetTargetLandmarks()->GetPoint(i, newPoint);
    ctNewPoints->InsertNextPoint(newPoint);
  }
  // END ICP TO CT
  */
  
  // ICP TO ERROR
  icp->SetSource(icpSource);
  icp->SetTarget(errorArtery); 
  icp->GetLandmarkTransform()->SetModeToRigidBody();
  icp->Modified();
  icp->Update();
  // std::cout << "The icp matrix is: " << *icp->GetMatrix() << std::endl; 
 
  vtkSmartPointer<vtkPoints> ctNewPoints =
    vtkSmartPointer<vtkPoints>::New();
  ctNewPoints->DeepCopy(ctReader->GetOutput()->GetPoints());
  vtkSmartPointer<vtkPoints> frameNewPoints =
    vtkSmartPointer<vtkPoints>::New();
  frameNewPoints->DeepCopy(someError->GetOutput()->GetPoints());
  double newPoint[3];

  // if success, SourceLandmarks == TargetLandmarks
  for (vtkIdType i = 0; i < noOfPoints; i++) {
    icpSource->GetPoint(i, newPoint);
    frameNewPoints->InsertNextPoint(newPoint);
    icp->GetLandmarkTransform()->GetTargetLandmarks()->GetPoint(i, newPoint);
    artery->GetPoint(errorArtery->FindPoint(newPoint), newPoint);
    ctNewPoints->InsertNextPoint(newPoint);
  }
  // END ICP TO ERROR
  
  // ctNewPoints->Print(std::cout);
  // frameNewPoints->Print(std::cout);

  // Re-registration: Get the corrected registration from ct to frame coordinates
  vtkSmartPointer<vtkLandmarkTransform> registerCorrected =
    vtkSmartPointer<vtkLandmarkTransform>::New();
  registerCorrected->SetSourceLandmarks(ctNewPoints);
  registerCorrected->SetTargetLandmarks(frameNewPoints);
  registerCorrected->SetModeToRigidBody();
  registerCorrected->Update();
  // registerCorrected->GetMatrix()->Print(std::cout);

  // Transform artery surface from ct to corrected frame coordinates
  vtkSmartPointer<vtkTransformPolyDataFilter> transformCorrected =
    vtkSmartPointer<vtkTransformPolyDataFilter>::New();
  transformCorrected->SetInputConnection(arteryReader->GetOutputPort());
  transformCorrected->SetTransform(registerCorrected);
  transformCorrected->Update();

  // Calculate average error from ct to error frame coordinates
  vtkSmartPointer<vtkPolyData> correctedArtery = transformCorrected->GetOutput();
  double correctedPoint[3];
  averageDistance = 0;
  for (vtkIdType i = 0; i < transformTrue->GetOutput()->GetPoints()->GetNumberOfPoints(); 
       i++) {
    trueArtery->GetPoint(i, truePoint);
    correctedArtery->GetPoint(i, correctedPoint);
    averageDistance += sqrt( (truePoint[0]-correctedPoint[0])*(truePoint[0]-correctedPoint[0]) +
                             (truePoint[1]-correctedPoint[1])*(truePoint[1]-correctedPoint[1]) +
                             (truePoint[2]-correctedPoint[2])*(truePoint[2]-correctedPoint[2]) );
    
  }
  float correctedError = averageDistance/transformTrue->GetOutput()->GetPoints()->GetNumberOfPoints();
  // std::cout << correctedError << std::endl;

  // Prep for visualization

  vtkSmartPointer<vtkPolyData> source = ctReader->GetOutput();
  vtkSmartPointer<vtkPolyData> target = frameReader->GetOutput();
  vtkSmartPointer<vtkPolyData> error = someError->GetOutput();

  vtkSmartPointer<vtkVertexGlyphFilter> sourceGlyphFilter =
    vtkSmartPointer<vtkVertexGlyphFilter>::New();
#if VTK_MAJOR_VERSION <= 5
  sourceGlyphFilter->SetInputConnection(source->GetProducerPort());
#else
  sourceGlyphFilter->SetInputData(source);
#endif
  sourceGlyphFilter->Update();
  
  vtkSmartPointer<vtkVertexGlyphFilter> targetGlyphFilter =
    vtkSmartPointer<vtkVertexGlyphFilter>::New();
#if VTK_MAJOR_VERSION <= 5
  targetGlyphFilter->SetInputConnection(target->GetProducerPort());
#else
  targetGlyphFilter->SetInputData(target);
#endif
  targetGlyphFilter->Update();

  vtkSmartPointer<vtkVertexGlyphFilter> errorGlyphFilter =
    vtkSmartPointer<vtkVertexGlyphFilter>::New();
  errorGlyphFilter->SetInputConnection(error->GetProducerPort());
  targetGlyphFilter->Update();

  vtkSmartPointer<vtkVertexGlyphFilter> icpSourceGlyphFilter =
    vtkSmartPointer<vtkVertexGlyphFilter>::New();
  icpSourceGlyphFilter->SetInputConnection(icpSource->GetProducerPort());
  targetGlyphFilter->Update();

  vtkSmartPointer<vtkPolyData> temp2 =
    vtkSmartPointer<vtkPolyData>::New();
  temp2->SetPoints(frameNewPoints);
  vtkSmartPointer<vtkVertexGlyphFilter> reregFrameGlyphFilter =
    vtkSmartPointer<vtkVertexGlyphFilter>::New();
  reregFrameGlyphFilter->SetInputConnection(temp2->GetProducerPort());
  reregFrameGlyphFilter->Update();

  // Transform fiducials in ct coordinates to fiducials in corrected frame coordinates
  vtkSmartPointer<vtkTransformPolyDataFilter> transformFidsCorrected =
    vtkSmartPointer<vtkTransformPolyDataFilter>::New();
  transformFidsCorrected->SetInputConnection(source->GetProducerPort());
  transformFidsCorrected->SetTransform(registerCorrected);
  transformFidsCorrected->Update();

  // Visualize
  vtkSmartPointer<vtkPolyDataMapper> sourceMapper =
    vtkSmartPointer<vtkPolyDataMapper>::New();
  sourceMapper->SetInputConnection(sourceGlyphFilter->GetOutputPort());
  vtkSmartPointer<vtkActor> sourceActor =
    vtkSmartPointer<vtkActor>::New();
  sourceActor->SetMapper(sourceMapper);
  sourceActor->GetProperty()->SetColor(0,0,0);
  sourceActor->GetProperty()->SetPointSize(4);

  vtkSmartPointer<vtkPolyDataMapper> targetMapper =
    vtkSmartPointer<vtkPolyDataMapper>::New();
  targetMapper->SetInputConnection(targetGlyphFilter->GetOutputPort());
  vtkSmartPointer<vtkActor> targetActor =
    vtkSmartPointer<vtkActor>::New();
  targetActor->SetMapper(targetMapper);
  targetActor->GetProperty()->SetColor(0,0,1);
  targetActor->GetProperty()->SetPointSize(4);
  
  vtkSmartPointer<vtkPolyDataMapper> ctArteryMapper =
    vtkSmartPointer<vtkPolyDataMapper>::New();
  ctArteryMapper->SetInputConnection(arteryReader->GetOutputPort());
  vtkSmartPointer<vtkActor> ctArteryActor =
    vtkSmartPointer<vtkActor>::New();
  ctArteryActor->SetMapper(ctArteryMapper);
  ctArteryActor->GetProperty()->SetColor(0,0,0);

  vtkSmartPointer<vtkPolyDataMapper> solutionMapper =
    vtkSmartPointer<vtkPolyDataMapper>::New();
  solutionMapper->SetInputConnection(transformTrue->GetOutputPort());
  vtkSmartPointer<vtkActor> solutionActor =
    vtkSmartPointer<vtkActor>::New();
  solutionActor->SetMapper(solutionMapper);
  solutionActor->GetProperty()->SetColor(0,0,1);

  vtkSmartPointer<vtkPolyDataMapper> errorMapper =
    vtkSmartPointer<vtkPolyDataMapper>::New();
  errorMapper->SetInputConnection(errorGlyphFilter->GetOutputPort());
  vtkSmartPointer<vtkActor> errorActor =
    vtkSmartPointer<vtkActor>::New();
  errorActor->SetMapper(errorMapper);
  errorActor->GetProperty()->SetColor(1,0,0);
  errorActor->GetProperty()->SetPointSize(4);
  
  vtkSmartPointer<vtkPolyDataMapper> icpSourceMapper =
    vtkSmartPointer<vtkPolyDataMapper>::New();
  icpSourceMapper->SetInputConnection(icpSourceGlyphFilter->GetOutputPort());
  vtkSmartPointer<vtkActor> icpSourceActor =
    vtkSmartPointer<vtkActor>::New();
  icpSourceActor->SetMapper(icpSourceMapper);
  icpSourceActor->GetProperty()->SetColor(1,0,0);
  icpSourceActor->GetProperty()->SetPointSize(4);

  vtkSmartPointer<vtkPolyDataMapper> frameErrorArteryMapper =
    vtkSmartPointer<vtkPolyDataMapper>::New();
  frameErrorArteryMapper->SetInputConnection(transformError->GetOutputPort());
  vtkSmartPointer<vtkActor> frameErrorArteryActor =
    vtkSmartPointer<vtkActor>::New();
  frameErrorArteryActor->SetMapper(frameErrorArteryMapper);
  frameErrorArteryActor->GetProperty()->SetColor(1,0,0);
  
  vtkSmartPointer<vtkPolyDataMapper> reregFrameMapper =
    vtkSmartPointer<vtkPolyDataMapper>::New();
  reregFrameMapper->SetInputConnection(reregFrameGlyphFilter->GetOutputPort());
  vtkSmartPointer<vtkActor> reregFrameActor =
    vtkSmartPointer<vtkActor>::New();
  reregFrameActor->SetMapper(reregFrameMapper);
  reregFrameActor->GetProperty()->SetColor(1,1,0);
  reregFrameActor->GetProperty()->SetPointSize(4);
  
  vtkSmartPointer<vtkPolyDataMapper> frameCorrectedArteryMapper =
    vtkSmartPointer<vtkPolyDataMapper>::New();
  frameCorrectedArteryMapper->SetInputConnection(transformCorrected->GetOutputPort());
  vtkSmartPointer<vtkActor> frameCorrectedArteryActor =
    vtkSmartPointer<vtkActor>::New();
  frameCorrectedArteryActor->SetMapper(frameCorrectedArteryMapper);
  frameCorrectedArteryActor->GetProperty()->SetColor(1,1,0);
   
  vtkSmartPointer<vtkPolyDataMapper> frameCorrectedFidsMapper =
    vtkSmartPointer<vtkPolyDataMapper>::New();
  frameCorrectedFidsMapper->SetInputConnection(transformFidsCorrected->GetOutputPort());
  vtkSmartPointer<vtkActor> frameCorrectedFidsActor =
    vtkSmartPointer<vtkActor>::New();
  frameCorrectedFidsActor->SetMapper(frameCorrectedFidsMapper);
  frameCorrectedFidsActor->GetProperty()->SetColor(1,1,0);
  frameCorrectedFidsActor->GetProperty()->SetPointSize(4);
  
  // Create a renderer, render window, and interactor
  vtkSmartPointer<vtkRenderer> renderer =
    vtkSmartPointer<vtkRenderer>::New();
  vtkSmartPointer<vtkRenderWindow> renderWindow =
    vtkSmartPointer<vtkRenderWindow>::New();
  renderWindow->AddRenderer(renderer);
  vtkSmartPointer<vtkRenderWindowInteractor> renderWindowInteractor =
    vtkSmartPointer<vtkRenderWindowInteractor>::New();
  renderWindowInteractor->SetRenderWindow(renderWindow);

  // Add the actor to the scene
  // renderer->AddActor(sourceActor);
  //renderer->AddActor(targetActor);
  // renderer->AddActor(ctArteryActor);
  renderer->AddActor(solutionActor); // green
  // renderer->AddActor(errorActor); // blue
  renderer->AddActor(icpSourceActor); // black
  // renderer->AddActor(frameErrorArteryActor); // blue
  // renderer->AddActor(reregFrameActor);
  // renderer->AddActor(frameCorrectedArteryActor); // yellow
  // renderer->AddActor(frameCorrectedFidsActor); //yellow
  renderer->SetBackground(0.3, 0.6, 0.3); // Background color green

  // Render and interact
  renderWindow->Render();
  renderWindowInteractor->Start();
   
  reportError[0] = errorError;
  reportError[1] = correctedError;
  return 1;
}

int main(int, char *[])
{
  srand(time(0));
  float reportError[2];
  for (int i = 0; i < 1; i++)
  {
    run(reportError);
    std::cout << reportError[1] << std::endl;
  }
  return EXIT_SUCCESS;
}
