/* -*- Mode: C++; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-    */
/* ex: set filetype=cpp softtabstop=4 shiftwidth=4 tabstop=4 cindent expandtab: */

/*
  $Id: mtsMedtronicStealthlinkExampleComponent.cpp 3836 2012-09-12 18:43:09Z dmirota1 $

  Author(s):  Peter Kazanzides, Anton Deguet
  Created on: 2011-07-14

  (C) Copyright 2011-2012 Johns Hopkins University (JHU), All Rights Reserved.

--- begin cisst license - do not edit ---

This software is provided "as is" under an open source license, with
no warranty.  The complete license can be found in license.txt and
http://www.cisst.org/cisst/license.txt.

--- end cisst license ---
*/

#include "mtsMedtronicStealthlinkExampleComponent.h"
#include <cisstMultiTask/mtsInterfaceRequired.h>
#include <cisstMultiTask/mtsTransformationTypes.h>
#include <cisstParameterTypes/prmPositionCartesianGet.h>

#ifndef SAW_MEDTRONIC_WITHOUT_STEALTHLINK_TYPES
    #ifdef SAW_MEDTRONIC_STEALTHLINK2
        #include <sawMedtronicStealthlink/mtsMedtronicStealthlink2Types.h>
    #else
        #include <sawMedtronicStealthlink/mtsMedtronicStealthlinkTypes.h>
    #endif
#endif

#include <cisstMultiTask/mtsFixedSizeVectorTypes.h>

mtsMedtronicStealthlinkExampleComponent::mtsMedtronicStealthlinkExampleComponent(const std::string & name,
                                                                                 double periodInSeconds):
    mtsTaskPeriodic(name, periodInSeconds),
    BatchReadyEventCounter(0),
    CollectionRunning(false),
    SamplesCollected(0)
{

   AddStealthlinkInterface();
}

mtsMedtronicStealthlinkExampleComponent::mtsMedtronicStealthlinkExampleComponent(const std::string & name,
                                                                                 double periodInSeconds,
                                                                                 bool enableStateCollectionInterface):
    mtsTaskPeriodic(name, periodInSeconds),
    BatchReadyEventCounter(0),
    CollectionRunning(false),
    SamplesCollected(0)
{

    AddStealthlinkInterface();

    //state collection
    if(enableStateCollectionInterface)
        AddStateCollectionInterface();
}

void mtsMedtronicStealthlinkExampleComponent::AddStealthlinkInterface()
{
    mtsInterfaceRequired * required = AddInterfaceRequired("Stealthlink");
    if (required) {
#ifndef SAW_MEDTRONIC_WITHOUT_STEALTHLINK_TYPES
        required->AddFunction("GetTool", Stealthlink.GetTool);
        required->AddFunction("GetFrame", Stealthlink.GetFrame);
#endif
    }

    // the following two tools are using names normally defined in config.xml
    AddToolInterface("Pointer", Pointer);
    AddToolInterface("Frame", Frame);

}

void mtsMedtronicStealthlinkExampleComponent::AddStateCollectionInterface()
{
    mtsInterfaceRequired * required;

    required = AddInterfaceRequired("CollectorState");
    if (required) {
        required->AddFunction("StartCollection", CollectorState.StartCollection);
        required->AddFunction("StopCollection", CollectorState.StopCollection);
    }
}

void mtsMedtronicStealthlinkExampleComponent::AddToolInterface(const std::string & toolName,
                                                               mtsMedtronicStealthlinkExampleComponent::ToolStruct & functionSet)
{
    mtsInterfaceRequired * required = AddInterfaceRequired(toolName, MTS_OPTIONAL);
    if (required) {
        required->AddFunction("GetPositionCartesian", functionSet.GetPositionCartesian);
        required->AddFunction("GetMarkerCartesian", functionSet.GetMarkerCartesian);
    }
}


void mtsMedtronicStealthlinkExampleComponent::Run(void)
{
#ifndef SAW_MEDTRONIC_WITHOUT_STEALTHLINK_TYPES
    mtsStealthTool StealthTool;
    mtsStealthFrame StealthFrame;
#endif
    prmPositionCartesianGet prmPos;
    mtsFrm3 mtsFrm;
    mtsDouble predictedAccuracy;

    bool debug = true;
    bool didOutput = false;
    bool getValid = false;
    mtsExecutionResult result;
#ifndef SAW_MEDTRONIC_WITHOUT_STEALTHLINK_TYPES
    bool getValid = false;
#endif // SAW_MEDTRONIC_WITHOUT_STEALTHLINK_TYPES

    std::cout << "hello!" << std::endl;

    if (Pointer.GetPositionCartesian.IsValid()) {
        result = Pointer.GetPositionCartesian(prmPos);
        if (!result.IsOK()) {
            std::cout << "Pointer.GetPositionCartesian() failed: " << result << std::endl;
        }
        if (prmPos.Valid()) {
            if(debug)
                std::cout << "Interface Pointer: " << prmPos.Position().Translation() << "; " << std::endl;
                std::cout << "Pointer Rotation: " << prmPos.Position().Rotation() << ";" << std::endl;
            didOutput = true;
        }
    }
    if (Pointer.GetMarkerCartesian.IsValid()) {
        result = Pointer.GetMarkerCartesian(prmPos);
        if (!result.IsOK()) {
            std::cout << "Pointer.GetMarkerCartesian() failed: " << result << std::endl;
        }
        if (prmPos.Valid()) {
            if(debug)
                std::cout << "Interface PointerM: " << prmPos.Position().Translation() << "; " << std::endl;
                std::cout << "Interface PointerM Rotation: " << prmPos.Position().Rotation() << std::endl;
            didOutput = true;
        }
    }

    if (Frame.GetPositionCartesian.IsValid()) {
        result = Frame.GetPositionCartesian(prmPos);
        if (!result.IsOK()) {
            std::cout << "Frame.GetPositionCartesian() failed: " << result << std::endl;
        }
        if (prmPos.Valid()) {
            if(debug)
                std::cout << "Interface Frame: " << prmPos.Position().Translation() << "; ";
            didOutput = true;
        } else {
            if(debug)
                std::cerr << "Interface Frame, invalid position" << std::endl;
            didOutput = true;
        }
    }
    if (Frame.GetMarkerCartesian.IsValid()) {
        result = Frame.GetMarkerCartesian(prmPos);
        if (!result.IsOK()) {
            std::cout << "Frame.GetMarkerCartesian() failed: " << result << std::endl;
        }
        if (prmPos.Valid()) {
            if(debug)
                std::cout << "Interface FrameM: " << prmPos.Position().Translation() << "; " << std::endl;
                std::cout << "FrameM Rotation: " << prmPos.Position().Rotation() << ";" << std::endl;
            didOutput = true;
        } else {
            if(debug)
                std::cerr << "Interface Frame, invalid position" << std::endl;
            didOutput = true;
        }
    }


    if (didOutput && debug) {
        std::cout << std::endl;
    } else {
        if(debug)
            std::cout << "." << std::flush;
    }
}
